//  RegImport.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include "windows.h"

#define BUFFER_SIZE		1024*16
#define DWORD_PREFIX	"dword:"
#define MULTISZ_PREFIX	"multi_sz:"
#define HEX_PREFIX		"hex:"

#define HKLM_LONG		"HKEY_LOCAL_MACHINE\\"
#define HKLM_SHORT		"HKLM\\"

#define HKCU_LONG		"HKEY_CURRENT_USER\\"
#define HKCU_SHORT		"HKCU\\"

#define HKU_LONG		"HKEY_USERS\\"
#define HKU_SHORT		"HKU\\"

#define HKCR_LONG		"HKEY_CLASSES_ROOT\\"
#define HKCR_SHORT		"HKCR\\"


void RemoveQuotes(char* str)
{
	if ((*str)=='\"')
	{
		str[strlen(str)-1]=0;
		memmove(str,str+1,strlen(str));
	}
}

HKEY OpenOrCreateKey(char* keyname)
{
	HKEY	root=NULL;
	DWORD	err;

	if (!strnicmp(keyname,HKLM_LONG,strlen(HKLM_LONG)))
	{
		root=HKEY_LOCAL_MACHINE;
		keyname+=strlen(HKLM_LONG);
	}
	else if (!strnicmp(keyname,HKLM_SHORT,strlen(HKLM_SHORT)))
	{
		root=HKEY_LOCAL_MACHINE;
		keyname+=strlen(HKLM_SHORT);
	}
	else if (!strnicmp(keyname,HKCU_LONG,strlen(HKCU_LONG)))
	{
		root=HKEY_CURRENT_USER;
		keyname+=strlen(HKCU_LONG);
	}
	else if (!strnicmp(keyname,HKCU_SHORT,strlen(HKCU_SHORT)))
	{
		root=HKEY_CURRENT_USER;
		keyname+=strlen(HKCU_SHORT);
	}
	else if (!strnicmp(keyname,HKU_LONG,strlen(HKU_LONG)))
	{
		root=HKEY_USERS;
		keyname+=strlen(HKU_LONG);
	}
	else if (!strnicmp(keyname,HKU_SHORT,strlen(HKU_SHORT)))
	{
		root=HKEY_USERS;
		keyname+=strlen(HKU_SHORT);
	}
	else if (!strnicmp(keyname,HKCR_LONG,strlen(HKCR_LONG)))
	{
		root=HKEY_CLASSES_ROOT;
		keyname+=strlen(HKCR_LONG);
	}
	else if (!strnicmp(keyname,HKCR_SHORT,strlen(HKCR_SHORT)))
	{
		root=HKEY_CLASSES_ROOT;
		keyname+=strlen(HKCR_SHORT);
	}

	if (root==NULL)
		return NULL;

	wchar_t unicodekeyname[MAX_PATH+1];

	MultiByteToWideChar(CP_UTF8,0,keyname,-1,unicodekeyname,MAX_PATH);

	HKEY key;

	if (err=RegCreateKeyEx(root,unicodekeyname,0,NULL,0,KEY_ALL_ACCESS,NULL,&key,NULL))
	{
		SetLastError(err);
		return NULL;
	}	
	return key;
}


// buffer used to read the file
char	buffer[BUFFER_SIZE+1];

// buffer used for conversion operations
wchar_t convbuffer[BUFFER_SIZE+1];


int wmain(int argc, wchar_t *argv[])
{
	if (argc!=2)
	{
		printf("Usage:\r\nregimport regfile\r\nregfile must be an ASCII text file in Windows registry format.\r\n");
		return -1;
	}

	HANDLE regfile=CreateFile(argv[1],GENERIC_READ,0,NULL,OPEN_EXISTING,0,NULL);

	if (regfile==INVALID_HANDLE_VALUE)
	{
		printf("Error opening input file.\r\n");
		return -1;
	}

	if (GetFileSize(regfile,NULL)>BUFFER_SIZE)
	{
		printf("Input file is too big. Maximum supported size is %d bytes.\r\n",BUFFER_SIZE);
		CloseHandle(regfile);
		return -1;
	}

	DWORD count;

	if (!ReadFile(regfile,buffer,BUFFER_SIZE,&count,NULL))
	{
		printf("Error %d reading input file.\r\n",GetLastError());
		CloseHandle(regfile);
		return -1;
	}

	buffer[count]=0;

	CloseHandle(regfile);

	// removes all the whitespaces
	DWORD	dest=0;
	BOOL	quotes=FALSE;
	BOOL	skiptoeol=FALSE;
	

	for (DWORD i=0;i<count;i++)
	{
		BOOL skip=FALSE;

		switch(buffer[i])
		{
		case ' ':
		case '\t':
			skip=quotes?FALSE:TRUE;
			break;
		case '\r':
			skip=TRUE;
			break;
		case '\n':
			quotes=FALSE;
			skiptoeol=FALSE;
			break;
		case '\\':
			if (quotes)
			{
				//skip current char and copy the next one
				i++;
			}
			break;
		case '\"':
			if (!skiptoeol)
				quotes=quotes?FALSE:TRUE;
			break;
		case ';':
			if (!quotes)
				skiptoeol=TRUE;
		}

		if ((skip)||(skiptoeol)) 
			continue;

		buffer[dest++]=buffer[i];
	}

	count=dest;
	buffer[count]=0;

	// parse the text row-by-row
	char keyname[MAX_PATH+1];

	DWORD rowindex=1;

	for  (char* row=strtok(buffer,"\n");row;row=strtok(NULL,"\n"),rowindex++)
	{
		// empty row
		if (*row==0)
			continue;

		// comment, we can skip it
		if (*row==';')
			continue;

		// regedit header, we can ignore it
		if ((!strcmpi(row,"REGEDIT4"))||(!strcmpi(row,"Windows Registry Editor Version 5.00")))
			continue;

		//  key name
		if (*row=='[')
		{
			strncpy(keyname,row+1,min(strlen(row),MAX_PATH));
			keyname[MAX_PATH]=0;
			if (keyname[strlen(keyname)-1]!=']')
			{
				printf("Keyname syntax error around line %d.\r\n",rowindex);
				return -1;
			}

			keyname[strlen(keyname)-1]=0;
			continue;
		}

		// split name and value
		char* value=strrchr(row,'=');

		if (value==NULL)
		{
			printf("Syntax error around line %d.\r\n",rowindex);
			return -1;
		}

		char* name=row;

		*value=0;
		value++;

		RemoveQuotes(name);

		HKEY	key;

		key=OpenOrCreateKey(keyname);

		if (key==NULL)
		{
			printf("Error %d opening key %s.\r\n",GetLastError(),keyname);
			return -1;
		}

		DWORD err;

		wchar_t* unicodename;
		wchar_t unicodebuffer[MAX_PATH+1];

		if (!strcmp(name,"@"))
		{
			unicodename=NULL;
		}
		else
		{
			MultiByteToWideChar(CP_UTF8,0,name,-1,unicodebuffer,MAX_PATH);

			unicodename=unicodebuffer;
		}

		if (!strnicmp(value,DWORD_PREFIX,strlen(DWORD_PREFIX)))
		{
			value+=strlen(DWORD_PREFIX);
			RemoveQuotes(value);

			// dword value
			char* endptr;

			DWORD val=strtoul(value,&endptr,16);

			if (*endptr!=0)
			{
				printf("Invalid integer value at row %d.\r\n",rowindex);
				RegCloseKey(key);
				return -1;
			}

			err=RegSetValueEx(key,unicodename,0,REG_DWORD,(BYTE*)&val,sizeof(DWORD));			
		}
		else if (!strnicmp(value,HEX_PREFIX,strlen(HEX_PREFIX)))
		{
			value+=strlen(HEX_PREFIX);
			RemoveQuotes(value);

			// hex value
			BYTE* vector=(BYTE*)convbuffer;
			DWORD index=0;

			for (char* v=strtok(value,",");v;v=strtok(NULL,","))
			{
				char* endptr;

				DWORD val=strtoul(v,&endptr,16);

				if ((*endptr!=0)||(val>255))
				{
					printf("Invalid byte value at row %d.\r\n",rowindex);
					RegCloseKey(key);
					return -1;
				}

				vector[index++]=(BYTE)val;
			}

			err=RegSetValueEx(key,unicodename,0,REG_BINARY,(BYTE*)vector,index);			
		}
		else if (!strnicmp(value,MULTISZ_PREFIX,strlen(MULTISZ_PREFIX)))
		{
			value+=strlen(MULTISZ_PREFIX);

			// multisz value
			char* p=value;
			char *q=value;
			BOOL quotes=FALSE;

			wchar_t* ptr=convbuffer;

			for (char* q=value;*q;q++)
			{
				switch (*q)
				{
				case '\\':
					//skip next char
					q++;
					break;
				case '\"':
					quotes=quotes?FALSE:TRUE;
					break;
				case ',':
					if (!quotes)
					{
						*q=0;
						q++;

						RemoveQuotes(p);

						ptr+=MultiByteToWideChar(CP_UTF8,0,p,-1,ptr,BUFFER_SIZE);

						p=q;

						// the first char will be skipped
						if (*q=='\"')
							quotes=TRUE;
					}
					break;
				}
			}

			RemoveQuotes(p);
			ptr+=MultiByteToWideChar(CP_UTF8,0,p,-1,ptr,BUFFER_SIZE);
		
			*ptr=0;

			err=RegSetValueEx(key,unicodename,0,REG_MULTI_SZ,(BYTE*)convbuffer,((DWORD)ptr-(DWORD)convbuffer)+2);			
		}
		else
		{
			RemoveQuotes(value);
	
			// string value
			DWORD len=MultiByteToWideChar(CP_UTF8,0,value,-1,convbuffer,BUFFER_SIZE);

			err=RegSetValueEx(key,unicodename,0,REG_SZ,(BYTE*)convbuffer,len*sizeof(wchar_t));			
		}			

		RegCloseKey(key);		
	
		if (err)
		{
			printf("Error %d writing value %s under key %s.\r\n",err,name,keyname);
			return -1;
		}
	}

	RegFlushKey(HKEY_LOCAL_MACHINE);
	RegFlushKey(HKEY_CURRENT_USER);
	RegFlushKey(HKEY_CLASSES_ROOT);
	RegFlushKey(HKEY_USERS);
	return 0;
}
